from badger_utils.systems.aave_system import AaveSystem
from badger_utils.systems.compound_system import CompoundSystem
from badger_utils.systems.token_system import TokenSystem
from badger_utils.systems.yearn_system import YearnSystem


class ChainRegistry:
    def __init__(
        self,
        curve=None,
        convex=None,
        uniswap=None,
        open_zeppelin=None,
        aragon=None,
        sushiswap=None,
        sushi=None,
        gnosis_safe=None,
        onesplit=None,
        pickle=None,
        harvest=None,
        tokens=None,
        whales=None,
        multicall=None,
        multisend=None,
        pancake=None,
        badger=None,
        yearn=None,
        aave=None,
        compound=None,
        chainlink=None,
        defidollar=None,
        digg=None,
    ):
        self.curve = curve
        self.convex = convex
        self.uniswap = uniswap
        self.open_zeppelin = open_zeppelin
        self.aragon = aragon
        self.sushiswap = sushiswap
        self.sushi = sushi
        self.gnosis_safe = gnosis_safe
        self.onesplit = onesplit
        self.pickle = pickle
        self.harvest = harvest
        self.tokens = tokens
        self.whales = whales
        self.multicall = multicall
        self.multisend = multisend
        self.pancake = pancake
        self.badger = badger
        self.yearn = yearn
        self.aave = aave
        self.compound = compound
        self.defidollar = defidollar
        self.chainlink = chainlink
        self.digg = digg

    def yearn_system(self) -> YearnSystem:
        if self.yearn is None:
            raise Exception("No yearn system registered")
        return YearnSystem(self)

    def token_system(self) -> TokenSystem:
        if self.tokens is None:
            raise Exception("No yearn system registered")
        return TokenSystem(self)

    def aave_system(self) -> AaveSystem:
        if self.aave is None:
            raise Exception("No aave system registered")
        return AaveSystem(self)

    def compound_system(self) -> CompoundSystem:
        if self.aave is None:
            raise Exception("No aave system registered")
        return CompoundSystem(self)



from typing import Tuple

from brownie import accounts
from brownie import chain
from brownie import interface
from brownie import web3

from badger_utils.constants import MaxUint256


class SushiswapSystem:
    def __init__(self):
        from badger_utils.registry import registry
        self.contract_registry = registry.sushiswap
        self.factory = interface.IUniswapV2Factory(
            web3.toChecksumAddress(self.contract_registry.factory)
        )
        self.router = interface.IUniswapRouterV2(
            web3.toChecksumAddress(self.contract_registry.router)
        )
        self.chef = interface.ISushiChef(self.contract_registry.sushiChef)
        self.bar = interface.IxSushi(self.contract_registry.xsushiToken)

    def add_max_liquidity(self, token_a, token_b, signer):
        token_a = interface.IERC20(token_a)
        token_b = interface.IERC20(token_b)

        balance_a = token_a.balanceOf(signer) // 2
        balance_b = token_b.balanceOf(signer) // 2

        assert balance_a > 0
        assert balance_b > 0

        token_a.approve(self.router, MaxUint256, {"from": signer})
        token_b.approve(self.router, MaxUint256, {"from": signer})

        return self.router.addLiquidity(
            token_a.address,
            token_b.address,
            balance_a,
            balance_b,
            0,
            0,
            signer,
            chain.time() + 1000,
            {"from": signer},
        )

    def add_chef_rewards(self, pool):
        chef = self.chef

        owner = accounts.at(self.chef.owner(), force=True)
        # Make an average allocation of lp tokens.
        avg_alloc_point = chef.totalAllocPoint() / chef.poolLength()

        # Add staking pool (for rewards) if not exists or set the current allocation
        # to be receive average allocation points.
        pid, exists = self._get_pool(pool)
        if exists:
            chef.set(avg_alloc_point, pool, True, {"from": owner})
        else:
            # NB: If the lp token is added more than once, rewards will get messed up.
            chef.add(avg_alloc_point, pool, True, {"from": owner})

        pid = chef.poolLength() - 1
        chain.mine()

        chef.updatePool(pid, {"from": owner})
        chain.mine()

        return pid

    def _get_pool(self, pool) -> Tuple[int, bool]:
        chef = self.chef
        # Iterate over pools and look for pool first
        # NB: THIS IS EXPENSIVE AND SHOULD ONLY BE USED FOR TESTING.
        for pid in range(0, chef.poolLength()):
            (address, _, _, _) = chef.poolInfo(pid)
            if address == pool.address:
                return pid, True
        return -1, False



from brownie import BadgerRegistry  # noqa

from badger_utils.registry.base_registry import BaseRegistry
from badger_utils.registry.bsc_registry import bsc_registry
from badger_utils.registry.eth_registry import eth_registry


class ContractRegistries(BaseRegistry):
    """
    Contract registry for each chain
    """

    def __init__(self):
        super().__init__()
        self.registries["eth"] = eth_registry
        self.registries["development"] = eth_registry
        self.registries["bsc"] = bsc_registry


registries = ContractRegistries()
registry = registries.get_active_chain_registry()



from brownie import interface


class TokenSystem:
    def __init__(self, registry):
        self.registry = registry

    def erc20_by_key(self, key):
        if key not in self.registry.tokens:
            raise Exception(f"Token with key {key} not found in registry")

        address = self.registry.tokens[key]
        return self.erc20_by_address(address)

    def erc20_by_address(self, address):
        return interface.IERC20(address)



import os
from typing import Dict
from typing import Optional

import requests


ETHPLORER_API_KEY = os.environ.get("ETHPLORER_KEY") or "freekey"


def get_top_token_holders(token_address: str, limit: Optional[int] = 50) -> Dict:
    holders_response = requests.get(
        f"https://api.ethplorer.io/getTopTokenHolders/{token_address}",
        params={"apiKey": ETHPLORER_API_KEY, "limit": str(limit)},
    )
    holders_response.raise_for_status()
    return holders_response.json()



from typing import Dict
from typing import Optional
from typing import Union

import requests
from brownie import web3
from requests import HTTPError

from badger_utils.registry import registry


TOKEN_TO_ID_MAP = {
    web3.toChecksumAddress(registry.tokens.wbtc): "wrapped-bitcoin",
    web3.toChecksumAddress(registry.tokens.badger): "badger-dao",
    web3.toChecksumAddress(registry.tokens.digg): "digg",
    web3.toChecksumAddress(registry.tokens.cvx): "convex-finance",
    web3.toChecksumAddress(registry.tokens.crv): "curve-dao-token",
    web3.toChecksumAddress(registry.tokens.cvxCrv): "convex-crv",
    web3.toChecksumAddress(registry.tokens.sushi): "sushi",
    web3.toChecksumAddress(registry.tokens.xSushi): "xsushi",
    web3.toChecksumAddress(registry.tokens.bbadger): "badger-sett-badger",
    web3.toChecksumAddress(registry.tokens.spell): "spell-token",
    web3.toChecksumAddress(registry.tokens.renbtc): "renbtc",
    web3.toChecksumAddress(registry.tokens.sbtc): "sbtc",
    web3.toChecksumAddress(registry.tokens.tbtc): "tbtc",
    web3.toChecksumAddress(registry.tokens.bbtc): "binance-wrapped-btc",
    web3.toChecksumAddress(registry.tokens.obtc): "boringdao-btc",
    web3.toChecksumAddress(registry.tokens.pbtc): "ptokens-btc",
    web3.toChecksumAddress(registry.tokens.hbtc): "huobi-btc",
    web3.toChecksumAddress(registry.tokens.meta): "meta",
    web3.toChecksumAddress(registry.tokens.bnb): "binance-coin",
    web3.toChecksumAddress(registry.tokens.matic): "polygon",
    web3.toChecksumAddress(registry.tokens.aave): "aave",
    web3.toChecksumAddress(registry.tokens.comp): "compound",
    web3.toChecksumAddress(registry.tokens.ohm): "olympus",
    web3.toChecksumAddress(registry.tokens.swpr): "swapr",
}


def address_to_id(token_address: str) -> Union[str, bool]:
    checksummed = web3.toChecksumAddress(token_address)
    return TOKEN_TO_ID_MAP.get(checksummed, False)


def fetch_usd_value(token_address: str, amount: int) -> float:
    price = fetch_usd_price(token_address)
    return price * amount


def fetch_daily_twap(token_address: str) -> Optional[Dict]:
    response = requests.get(
        f"https://api.coingecko.com/api/v3/coins/{address_to_id(token_address)}",
        params="tickers=true&community_data=false&developer_data=false&sparkline=false"
    )
    try:
        response.raise_for_status()
    except HTTPError:
        return
    data = response.json()
    market_data = data["market_data"]
    return market_data


def fetch_usd_price(token_address: str) -> Optional[float]:
    response = requests.get(
        f"https://api.coingecko.com/api/v3/coins/{address_to_id(token_address)}",
        params="tickers=false&community_data=false&developer_data=false&sparkline=false"
    )
    try:
        response.raise_for_status()
    except HTTPError:
        return
    data = response.json()
    usd_price = data["market_data"]["current_price"]["usd"]
    return usd_price


def fetch_usd_price_eth() -> Optional[float]:
    response = requests.get(
        "https://api.coingecko.com/api/v3/coins/ethereum",
        params="?tickers=false&community_data=false&developer_data=false&sparkline=false"
    )
    try:
        response.raise_for_status()
    except HTTPError:
        return
    data = response.json()
    usd_price = data["market_data"]["current_price"]["usd"]
    return usd_price



from brownie.network import web3
from dotmap import DotMap

from badger_utils.constants import BBADGER
from badger_utils.constants import SPELL
from badger_utils.constants import WhaleRegistryAction
from badger_utils.registry.chain_registry import ChainRegistry
from badger_utils.registry.yearn_registry import YearnRegistry

aragon_registry = DotMap(
    addresses=DotMap(
        agentImpl="0x3a93c17fc82cc33420d1809dda9fb715cc89dd37",
        companyTemplate="0xd737632caC4d039C9B0EEcc94C12267407a271b5",
    )
)

gnosis_safe_registry = DotMap(
    addresses=DotMap(
        proxyFactory="0x76E2cFc1F5Fa8F6a5b3fC4c8F4788F0116861F9B",
        masterCopy="0x34CfAC646f301356fAa8B21e94227e3583Fe3F5F",
    )
)

onesplit_registry = (DotMap(contract="0x50FDA034C0Ce7a8f7EFDAebDA7Aa7cA21CC1267e"),)

uniswap_registry = DotMap(
    routerV2="0x7a250d5630B4cF539739dF2C5dAcb4c659F2488D",
    factoryV2="0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f",
    uniToken="0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
    uniStakingRewards=DotMap(eth_wbtc="0xa1484C3aa22a66C62b77E0AE78E15258bd0cB711"),
)

multicall = "0xeefba1e63905ef1d7acba5a8513c70307c1ce441"
multisend = "0x8D29bE29923b68abfDD21e541b9374737B49cdAD"

compound_registry = DotMap(
    comptroller=web3.toChecksumAddress("0x3d9819210a31b4961b30ef54be2aed79b9c9cd3b"),
    cTokens=DotMap(usdc="0x39AA39c021dfbaE8faC545936693aC917d5E7563"),
)

multichain_registry = DotMap(eth_address="0xC564EE9f21Ed8A2d8E7e76c085740d5e4c5FaFbE")

harvest_registry = DotMap(
    symbol="FARM",
    badgerTree="0x06466a741094f51b45FB150c6D1e857B3E879967",
    farmToken="0xa0246c9032bC3A600820415aE600c6388619A14D",
    depositHelper="0xf8ce90c2710713552fb564869694b2505bfc0846",
    vaults=DotMap(renCrv="0x9aa8f427a17d6b0d91b6262989edc7d45d6aedf8"),
    farms=DotMap(
        fWBtc="0x917d6480ec60cbddd6cbd0c8ea317bcc709ea77b",
        fRenCrv="0xa3cf8d1cee996253fad1f8e3d68bdcba7b3a3db5",
        fRenCrv2="0x5365A2C47b90EE8C9317faC20edC3ce7037384FB",
        farm="0xae024F29C26D6f71Ec71658B1980189956B0546D",
    ),
)

pickle_registry = DotMap(
    pickleToken="0x429881672B9AE42b8EbA0E26cD9C73711b891Ca5",
    pickleChef="0xbD17B1ce622d73bD438b9E658acA5996dc394b0d",
    jars=DotMap(renCrv="0x2E35392F4c36EBa7eCAFE4de34199b2373Af22ec"),
    pids=DotMap(uniPickleEth=0, pRenCrv=13),
    farms=DotMap(wethStaking="0xa17a8883dA1aBd57c690DF9Ebf58fC194eDAb66F"),
)

sushi_registry = DotMap(
    sushiToken="0x6b3595068778dd592e39a122f4f5a5cf09c90fe2",
    xsushiToken="0x8798249c2E607446EfB7Ad49eC89dD1865Ff4272",
    symbol="SUSHI",
    symbol_xsushi="XSUSHI",
    sushiChef="0xc2EdaD668740f1aA35E4D8f227fB8E17dcA888Cd",
    router="0xd9e1cE17f2641f24aE83637ab66a2cca9C378B9F",
    factory="0xC0AEe478e3658e2610c5F7A4A2E1777cE9e4f2Ac",
    lpTokens=DotMap(
        sushiBadgerWBtc="0x110492b31c59716AC47337E616804E3E3AdC0b4a",
        sushiWbtcWeth="0xCEfF51756c56CeFFCA006cD410B03FFC46dd3a58",
    ),
    pids=DotMap(sushiBadgerWBtc=73, sushiEthWBtc=21),
)

yearn_registry = YearnRegistry(
    registry="0x50c1a2ea0a861a967d9d0ffe2ae4012c2e053804",
    experimental_vaults={"wbtc": "0xA696a63cc78DfFa1a63E9E50587C197387FF6C7E"},
)

aave_registry = DotMap(lendingPoolV2="0x7d2768de32b0b80b7a3454c06bdac94a69ddc7a9")

yearn_registry = (
    DotMap(
        yvWBTC="0xcB550A6D4C8e3517A939BC79d0c7093eb7cF56B5",
    ),
)

convex_registry = DotMap(
    cvxHelperVault="0x53c8e199eb2cb7c01543c137078a038937a68e40",
    cvxCrvHelperVault="0x2B5455aac8d64C14786c3a29858E43b5945819C0",
)

curve_registry = DotMap(
    minter="0xd061D61a4d941c39E5453435B6345Dc261C2fcE0",
    crvToken="0xD533a949740bb3306d119CC777fa900bA034cd52",
    symbol="CRV",
    pools=DotMap(
        sbtcCrv=DotMap(
            swap="0x7fC77b5c7614E1533320Ea6DDc2Eb61fa00A9714",
            token="0x075b1bb99792c9E1041bA13afEf80C91a1e70fB3",
            gauge="0x705350c4BcD35c9441419DdD5d2f097d7a55410F",
        ),
        renCrv=DotMap(
            swap="0x93054188d876f558f4a66B2EF1d97d16eDf0895B",
            token="0x49849C98ae39Fff122806C06791Fa73784FB3675",
            gauge="0xB1F2cdeC61db658F091671F5f199635aEF202CAC",
        ),
        tbtcCrv=DotMap(
            swap="0xaa82ca713d94bba7a89ceab55314f9effeddc78c",
            # swap="0xC25099792E9349C7DD09759744ea681C7de2cb66",
            token="0x64eda51d3Ad40D56b9dFc5554E06F94e1Dd786Fd",
            gauge="0x6828bcF74279eE32f2723eC536c22c51Eed383C6",
        ),
        hbtcCrv=DotMap(
            token="0xb19059ebb43466C323583928285a49f558E572Fd",
            gauge="0x4c18E409Dc8619bFb6a1cB56D114C3f592E0aE79",
            swap="0x4CA9b3063Ec5866A4B82E437059D2C43d1be596F",
        ),
        pbtcCrv=DotMap(
            token="0xDE5331AC4B3630f94853Ff322B66407e0D6331E8",
            gauge="0xd7d147c6Bb90A718c3De8C0568F9B560C79fa416",
            swap="0x11F419AdAbbFF8d595E7d5b223eee3863Bb3902C",
        ),
        obtcCrv=DotMap(
            token="0x2fE94ea3d5d4a175184081439753DE15AeF9d614",
            gauge="0x11137B10C210b579405c21A07489e28F3c040AB1",
            swap="0xd5BCf53e2C81e1991570f33Fa881c49EEa570C8D",
        ),
        bbtcCrv=DotMap(
            token="0x410e3E86ef427e30B9235497143881f717d93c2A",
            gauge="0xdFc7AdFa664b08767b735dE28f9E84cd30492aeE",
            swap="0xC45b2EEe6e09cA176Ca3bB5f7eEe7C47bF93c756",
        ),
        triCrypto=DotMap(
            token="0xca3d75ac011bf5ad07a98d02f18225f9bd9a6bdf",
            swap="0x80466c64868E1ab14a1Ddf27A676C3fcBE638Fe5",
            gauge="0x331aF2E331bd619DefAa5DAc6c038f53FCF9F785",
        ),
        triCryptoDos=DotMap(
            token="0xc4AD29ba4B3c580e6D59105FFf484999997675Ff",
            swap="0xD51a44d3FaE010294C616388b506AcdA1bfAAE46",
            gauge="0x3993d34e7e99Abf6B6f367309975d1360222D446",
        ),
    ),
    pids=DotMap(
        renCrv=6,
        sbtcCrv=7,
        tbtcCrv=16,
        hbtcCrv=8,
        pbtcCrv=18,
        obtcCrv=20,
        bbtcCrv=19,
        triCrypto=37,
        triCryptoDos=38,
    ),
)

chainlink_registry = DotMap(btc_usd="0xF4030086522a5bEEa4988F8cA5B36dbC97BeE88c")


defidollar_registry = DotMap(
    addresses=DotMap(
        badgerSettPeak="0x41671BA1abcbA387b9b2B752c205e22e916BE6e3",
        core="0x2A8facc9D49fBc3ecFf569847833C380A13418a8",
    ),
    pools=[
        DotMap(
            id=0,
            sett="0x6dEf55d2e18486B9dDfaA075bc4e4EE0B28c1545",
        ),
        DotMap(
            id=1,
            sett="0xd04c48A53c111300aD41190D63681ed3dAd998eC",
        ),
        DotMap(
            id=2,
            sett="0xb9D076fDe463dbc9f915E5392F807315Bf940334",
        ),
    ],
)
badger_registry = DotMap(
    token="0x3472a5a71965499acd81997a54bba8d852c6e53d", symbol="BADGER"
)

digg_registry = DotMap(
    token="0x798D1bE841a82a273720CE31c822C61a67a601C3", symbol="DIGG"
)

eth_registry = ChainRegistry(
    curve=curve_registry,
    uniswap=uniswap_registry,
    aragon=aragon_registry,
    sushiswap=sushi_registry,
    sushi=sushi_registry,
    gnosis_safe=gnosis_safe_registry,
    pickle=pickle_registry,
    harvest=harvest_registry,
    multicall=multicall,
    multisend=multisend,
    badger=badger_registry,
    yearn=yearn_registry,
    aave=aave_registry,
    chainlink=chainlink_registry,
    compound=compound_registry,
    defidollar=defidollar_registry,
    digg=digg_registry,
    convex=convex_registry,
)

eth_registry.tokens = DotMap(
    weth="0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
    bbadger=BBADGER,
    spell=SPELL,
    wbtc="0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
    sbtc=web3.toChecksumAddress("0xfe18be6b3bd88a2d2a7f928d00292e7a9963cfc6"),
    matic=web3.toChecksumAddress("0x7d1afa7b718fb893db30a3abc0cfc608aacfebb0"),
    aave=web3.toChecksumAddress("0x7fc66500c84a76ad7e9c93437bfc5ac33e2ddae9"),
    comp=web3.toChecksumAddress("0xc00e94cb662c3520282e6f5717214004a7f26888"),
    ohm=web3.toChecksumAddress("0x64aa3364f17a4d01c6f1751fd97c2bd3d7e7f1d5"),
    swpr=web3.toChecksumAddress("0x6cacdb97e3fc8136805a9e7c342d866ab77d0957"),
    crv="0xD533a949740bb3306d119CC777fa900bA034cd52",
    tbtc=web3.toChecksumAddress("0x8daebade922df735c38c80c7ebd708af50815faa"),
    bnb=web3.toChecksumAddress("0xB8c77482e45F1F44dE1745F52C74426C631bDD52"),
    bbtc=web3.toChecksumAddress("0x9be89d2a4cd102d8fecc6bf9da793be995c22541"),
    obtc=web3.toChecksumAddress("0x8064d9ae6cdf087b1bcd5bdf3531bd5d8c537a68"),
    pbtc=web3.toChecksumAddress("0x5228a22e72ccc52d415ecfd199f99d0665e7733b"),
    hbtc=web3.toChecksumAddress("0x0316eb71485b0ab14103307bf65a021042c6d380"),
    meta=web3.toChecksumAddress("0xa3bed4e1c75d00fa6f4e5e6922db7261b5e9acd2"),
    usdt=web3.toChecksumAddress("0xdac17f958d2ee523a2206206994597c13d831ec7"),
    badger=eth_registry.badger.token,
    digg="0x798D1bE841a82a273720CE31c822C61a67a601C3",
    farm=eth_registry.harvest.farmToken,
    sushi=eth_registry.sushi.sushiToken,
    xSushi=eth_registry.sushi.xsushiToken,
    usdc=web3.toChecksumAddress("0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48"),
    renbtc=web3.toChecksumAddress("0xeb4c2781e4eba804ce9a9803c67d0893436bb27d"),
    usdp=web3.toChecksumAddress("0x1456688345527bE1f37E9e627DA0837D6f08C925"),
    ibbtc=web3.toChecksumAddress("0xc4E15973E6fF2A35cC804c2CF9D2a1b817a8b40F"),
    dfd=web3.toChecksumAddress("0x20c36f062a31865bed8a5b1e512d9a1a20aa333a"),
    ausdc="0xBcca60bB61934080951369a648Fb03DF4F96263C",
    cvx="0x4e3FBD56CD56c3e72c1403e103b45Db9da5B9D2B",
    cvxCrv="0x62B9c7356A2Dc64a1969e19C23e4f579F9810Aa7",
    pnt="0x89Ab32156e46F46D02ade3FEcbe5Fc4243B9AAeD",
    bor="0x3c9d6c1C73b31c837832c72E04D3152f051fc1A9",
)

eth_registry.whales = DotMap(
    badger=DotMap(
        whale="0x19d099670a21bC0a8211a89B84cEdF59AbB4377F",
        token="0x3472A5A71965499acd81997a54BBA8D852C6E53d",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bBadger=DotMap(
        whale="0xa9429271a28F8543eFFfa136994c0839E7d7bF77",
        token="0x19D97D8fA813EE2f51aD4B4e04EA08bAf4DFfC28",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    harvestSuperSett=DotMap(
        whale="0xeD0B7f5d9F6286d00763b0FFCbA886D8f9d56d5e",
        token="0xAf5A1DECfa95BAF63E0084a35c62592B774A2A87",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    uniBadgerWbtc=DotMap(
        whale="0x235c9e24D3FB2FAFd58a2E49D454Fdcd2DBf7FF1",
        token="0xcD7989894bc033581532D2cd88Da5db0A4b12859",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    uniDiggWbtc=DotMap(
        whale="0xc17078fdd324cc473f8175dc5290fae5f2e84714",
        token="0xe86204c4eddd2f70ee00ead6805f917671f56c52",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    sbtcCrv=DotMap(
        whale="0x282742940ee0b7ed028bb48052bb4922282234da",
        token=eth_registry.curve.pools.sbtcCrv.token,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bSbtcCrv=DotMap(
        whale="0x10fc82867013fce1bd624fafc719bb92df3172fc",
        token="0xd04c48A53c111300aD41190D63681ed3dAd998eC",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    renCrv=DotMap(
        whale="0xc280f35eeb97564dfd6bf80722e031d8f5bd82c9",
        token=eth_registry.curve.pools.renCrv.token,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bRenCrv=DotMap(
        whale="0x2296f174374508278dc12b806a7f27c87d53ca15",
        token="0x6dEf55d2e18486B9dDfaA075bc4e4EE0B28c1545",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    tbtcCrv=DotMap(
        whale="0xb65cef03b9b89f99517643226d76e286ee999e77",
        token=eth_registry.curve.pools.tbtcCrv.token,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bTbtcCrv=DotMap(
        whale="0x085a9340ff7692ab6703f17ab5ffc917b580a6fd",
        token="0xb9D076fDe463dbc9f915E5392F807315Bf940334",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    hbtcCrv=DotMap(
        whale="0xcc775989e76ab386e9253df5b0c0b473e22102e2",
        token=eth_registry.curve.pools.hbtcCrv.token,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    pbtcCrv=DotMap(
        whale="0x67031973f76abcd80d8635ee18865813298923b4",
        token=eth_registry.curve.pools.pbtcCrv.token,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    obtcCrv=DotMap(
        whale="0x966a70a4d3719a6de6a94236532a0167d5246c72",
        token=eth_registry.curve.pools.obtcCrv.token,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bbtcCrv=DotMap(
        whale="0x93a62da5a14c80f265dabc077fcee437b1a0efde",
        token=eth_registry.curve.pools.bbtcCrv.token,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    wbtc=DotMap(
        whale="0xc11b1268c1a384e55c48c2391d8d480264a3a7f4",
        token="0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    sushiBadgerWbtc=DotMap(
        whale="0xc2EdaD668740f1aA35E4D8f227fB8E17dcA888Cd",
        token="0x110492b31c59716AC47337E616804E3E3AdC0b4a",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    sushiDiggWbtc=DotMap(
        whale="0xd16fda96cb572da89e4e39b04b99d99a8e3071fb",
        token="0x110492b31c59716AC47337E616804E3E3AdC0b4a",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    sushiWbtcEth=DotMap(
        whale="0xc2EdaD668740f1aA35E4D8f227fB8E17dcA888Cd",
        token="0xCEfF51756c56CeFFCA006cD410B03FFC46dd3a58",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bSushiWbtcEth=DotMap(
        whale="0x032c701886ad0317f0e58c8f4a570c6f9c0bbf4a",
        token="0x758A43EE2BFf8230eeb784879CdcFF4828F2544D",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    usdc=DotMap(
        whale="0xbe0eb53f46cd790cd13851d5eff43d12404d33e8",  # binance
        token="0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    digg=DotMap(
        whale="0x4a8651F2edD68850B944AD93f2c67af817F39F62",
        token="0x798D1bE841a82a273720CE31c822C61a67a601C3",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    renbtc=DotMap(
        whale="0x35ffd6e268610e764ff6944d07760d0efe5e40e5",
        token="0xeb4c2781e4eba804ce9a9803c67d0893436bb27d",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    triCrypto=DotMap(
        whale="0x9f719e0bc35c46236b3f450852b526d84fed514b",
        token="0xcA3d75aC011BF5aD07a98d02f18225F9bD9A6BDF",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    cvx=DotMap(
        whale="0xdd5bc57bf90e6c6b341120e5b38fb6eda8e6481d",
        token="0x4e3fbd56cd56c3e72c1403e103b45db9da5b9d2b",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    cvxCrv=DotMap(
        whale="0xca078cb6eb4b80b4ef84751230c49c8ce076c6ff",
        token="0x62b9c7356a2dc64a1969e19c23e4f579f9810aa7",
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
)



from brownie import interface


class AaveSystem:
    def __init__(self, registry):
        self.registry = registry
        self.lendingPool = interface.ILendingPool(registry.aave.lendingPoolV2)

    def deposit(self, asset, amount, params):
        self.lendingPool.deposit(
            asset, amount, params["from"], 0, {"from": params["from"]}
        )



from badger_utils.constants import ARBITRUM_NETWORK
from badger_utils.constants import BADGER_REGISTRY_ADDRESS
from badger_utils.constants import ETHEREUM_NETWORK
from badger_utils.constants import DEVELOPMENT_NETWORK
from badger_utils.constants import POLYGON_NETWORK
from badger_utils.registry.base_registry import BaseRegistry


class OnChainContractRegistries(BaseRegistry):
    """
    Contract registry for each chain pulled from on-chain
    """

    def initialize(self):
        """
        Call this method to map BadgerRegistry contract to on-chain implementation
        """
        try:
            from brownie import BadgerRegistry  # noqa
        except ImportError:
            return
        # Badger registry has same address in all networks
        if not self.registries:
            for network in [ETHEREUM_NETWORK, POLYGON_NETWORK,
                            ARBITRUM_NETWORK, DEVELOPMENT_NETWORK]:
                self.registries[network] = BadgerRegistry.at(BADGER_REGISTRY_ADDRESS)


chain_registries = OnChainContractRegistries()



from datetime import datetime
from typing import Union

ONE_MINUTE = 60
ONE_HOUR = 3600
ONE_DAY = 24 * ONE_HOUR
ONE_YEAR = 1 * 365 * ONE_DAY


def days_to_seconds(amount_days: Union[int, float]) -> int:
    return int(amount_days * ONE_DAY)


def hours_to_seconds(amount_hours: Union[int, float]) -> int:
    return int(amount_hours * 3600.0)


def minutes_to_seconds(amount_minutes: Union[int, float]) -> int:
    return int(amount_minutes * 60.0)


def to_utc_date(timestamp: int) -> str:
    return datetime.utcfromtimestamp(timestamp).strftime("%Y-%m-%d %H:%M:%S")


def to_timestamp(date: datetime) -> int:
    print(date.timestamp())
    return int(date.timestamp())


def seconds_to_minutes(duration: int) -> Union[int, float]:
    return duration / ONE_MINUTE


def seconds_to_days(duration: int) -> Union[int, float]:
    return duration / ONE_DAY


def seconds_to_hours(duration: int) -> Union[int, float]:
    return duration / ONE_HOUR



class YearnRegistry:
    def __init__(self, registry=None, experimental_vaults=None):
        self.registry = registry
        self.experimental_vaults = experimental_vaults

    def get_experimental_vault(self, key):
        return self.experimental_vaults[key]



from time import sleep
from typing import List
from typing import Optional

from brownie import *
from brownie.network.account import Account

from badger_utils.constants import TARGET_TOKENS
from badger_utils.ethplorer_utils import get_top_token_holders
from badger_utils.token_utils.utils import distribute_test_ether
from badger_utils.utils import is_address_eoa


class TokenWhaleNotFound(Exception):
    pass


def distribute_from_whale_realtime(
        recipient: Account,
        percentage: Optional[float] = 0.2,
        token_address: Optional[str] = TARGET_TOKENS[0]) -> None:
    token_whales = get_top_token_holders(token_address)['holders']
    target_whale = ''
    for token_whale in token_whales:
        # Use first whale that is not a contract
        if is_address_eoa(web3.toChecksumAddress(token_whale['address'])):
            target_whale = token_whale['address']  # type: str
            break
    if not target_whale:
        raise TokenWhaleNotFound(f"Cannot find whale for token {token_address}")
    if recipient.balance() < 2 * 10 ** 18:
        distribute_test_ether(recipient, Wei("5 ether"))
    force_ether = ForceEther.deploy({"from": recipient})
    recipient.transfer(force_ether, Wei("2 ether"))
    force_ether.forceSend(target_whale, {"from": recipient})

    token = interface.IERC20(token_address)
    token.transfer(
        recipient,
        token.balanceOf(target_whale) * percentage,
        {"from": target_whale},
    )


def distribute_from_whales_realtime(
        recipient: Account,
        percentage: Optional[float] = 0.8,
        tokens: Optional[List[str]] = None) -> None:
    """
    NOTE: This only works on Ethereum, since Ethplorer doesn't support any other chain
    """
    if not tokens:
        tokens = TARGET_TOKENS
    # Normal Transfers
    for token_addr in tokens:
        distribute_from_whale_realtime(recipient, percentage=percentage, token_address=token_addr)
        # This is needed because Ethplrorer API will raise exc if requests are made too often
        # for the free API key
        sleep(0.5)



from typing import List

from brownie.network import web3
from brownie.network.account import Account
from brownie.network.contract import Contract
from brownie.network.gas.strategies import GasNowStrategy

from badger_utils.registry.artifacts import artifacts

gas_strategy = GasNowStrategy("rapid")
ADMIN_SLOT = int(0xB53127684A568B3173AE13B9F8A6016E243E63B6E8EE1178D6A717850B5D6103)

def deploy_proxy_admin(deployer: Account) -> Contract:
    abi = artifacts.open_zeppelin["ProxyAdmin"]["abi"]
    bytecode = artifacts.open_zeppelin["ProxyAdmin"]["bytecode"]

    proxy_admin = web3.eth.contract(abi=abi, bytecode=bytecode)

    deploy_txn = proxy_admin.constructor().buildTransaction()
    transaction = deployer.transfer(data=deploy_txn["data"])

    return Contract.from_abi(
        "ProxyAdmin",
        web3.toChecksumAddress(transaction.contract_address),
        abi,
    )


def deploy_proxy_uninitialized(
    contract_name: str, logic_abi: List, logic: str, proxy_admin: str, deployer: Account
) -> Contract:
    abi = artifacts.open_zeppelin["AdminUpgradeabilityProxy"]["abi"]
    bytecode = artifacts.open_zeppelin["AdminUpgradeabilityProxy"]["bytecode"]

    admin_upgradeability_proxy = web3.eth.contract(abi=abi, bytecode=bytecode)

    deploy_txn = admin_upgradeability_proxy.constructor(
        logic, proxy_admin, web3.toBytes(hexstr="0x")
    ).buildTransaction()

    transaction = deployer.transfer(data=deploy_txn["data"])

    return Contract.from_abi(contract_name, transaction.contract_address, logic_abi)


def deploy_proxy(
    contract_name: str, logic_abi: List, logic: str,
    proxy_admin: str, initializer: str, deployer: Account
):
    abi = artifacts.open_zeppelin["AdminUpgradeabilityProxy"]["abi"]
    bytecode = artifacts.open_zeppelin["AdminUpgradeabilityProxy"]["bytecode"]

    admin_upgradeability_proxy = web3.eth.contract(abi=abi, bytecode=bytecode)

    deploy_txn = admin_upgradeability_proxy.constructor(
        logic, proxy_admin, web3.toBytes(hexstr=initializer)
    ).buildTransaction()

    transaction = deployer.transfer(data=deploy_txn["data"])

    return Contract.from_abi(contract_name, transaction.contract_address, logic_abi)


def get_proxy_admin(proxy: str) -> str:
    return web3.toChecksumAddress(
        web3.eth.getStorageAt(proxy, ADMIN_SLOT).hex()
    )



from typing import Optional
from typing import Union

from brownie import interface

from badger_utils.constants import DIGG


class DiggUtils:
    def __init__(self, address: Optional[str] = DIGG):
        self.digg = interface.IDigg(address)
        self.sharesPerFragment = self.digg._sharesPerFragment()
        self.initialShares = self.digg._initialSharesPerFragment()

    def shares_to_fragments(self, shares: int) -> Union[int, float]:
        if shares == 0:
            return 0
        return self.sharesPerFragment / shares



from typing import Any
from typing import List
from typing import Optional
from typing import Union

from brownie import *
from brownie.network.account import Account

from badger_utils.constants import WhaleRegistryAction
from badger_utils.registry import registry
from badger_utils.token_utils.balances import Balances


def get_token_balances(tokens: List[Any], accounts_to_check: List[Account]) -> Balances:
    balances = Balances()
    for token in tokens:
        for account in accounts_to_check:
            balances.set(token, account, token.balanceOf(account))
    return balances


def distribute_from_whale(
        whale_config, recipient: Account,
        percentage: Optional[float] = 0.2) -> None:
    if whale_config.action == WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT:
        force_ether = ForceEther.deploy({"from": recipient})
        if recipient.balance() < 2 * 10 ** 18:
            distribute_test_ether(recipient, Wei("2 ether"))
        recipient.transfer(force_ether, Wei("2 ether"))
        force_ether.forceSend(whale_config.whale, {"from": recipient})

    token = interface.IERC20(whale_config.token)
    token.transfer(
        recipient,
        token.balanceOf(whale_config.whale) * percentage,
        {"from": whale_config.whale},
    )


def distribute_from_whales(
        recipient: Account,
        percentage: Optional[float] = 0.8, assets: Union[Optional[str], List[str]] = "All") -> None:

    accounts[0].transfer(recipient, Wei("50 ether"))
    # Normal Transfers
    for key, whale_config in registry.whales.items():
        if assets != "All" and key not in assets:
            continue
        # Handle special cases after all standard distributions
        if whale_config.special:
            continue
        if key != "_pytestfixturefunction":
            distribute_from_whale(whale_config, recipient, percentage=percentage)


def distribute_test_ether(recipient: Account, amount: int) -> None:
    """
    On test environments, transfer ETH from default ganache account to specified account
    """
    idx = 0
    while idx < len(accounts):
        if accounts[idx].balance() >= amount:
            break
        idx += 1
    if idx != len(accounts):
        accounts[idx].transfer(recipient, amount)



import requests
from brownie.network.gas.bases import SimpleGasStrategy


GASNOW_COMPATIBLE_API = "https://etherchain.org/api/gasnow"


# This module is a port for GasNowStrategy strategy. As GasNow service is closed, use
# gasnow org API which is a Gasnow API compatible enpoint
class GasNowCompatibleStrategy(SimpleGasStrategy):
    def __init__(self, speed: str = "fast"):
        if speed not in ("rapid", "fast", "standard", "slow"):
            raise ValueError("`speed` must be one of: rapid, fast, standard, slow")
        self.speed = speed

    def get_gas_price(self) -> int:
        response = requests.get(GASNOW_COMPATIBLE_API)
        response.raise_for_status()
        return response.json()["data"][self.speed]



import os

from dotmap import DotMap
import json


os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/aragon",
    "Agent.json",
)
with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/aragon",
    "Agent.json",
)) as f:
    Agent = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/aragon",
    "Vault.json",
)) as f:
    Vault = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/aragon",
    "Voting.json",
)) as f:
    Voting = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/aragon",
    "Finance.json",
)) as f:
    Finance = json.load(f)


with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/aragon",
    "TokenManager.json",
)) as f:
    TokenManager = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/aragon",
    "CompanyTemplate.json",
)) as f:
    CompanyTemplate = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/aragon",
    "MiniMeToken.json",
)) as f:
    MiniMeToken = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/gnosis-safe",
    "MasterCopy.json",
)) as f:
    MasterCopy = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/gnosis-safe",
    "ProxyFactory.json",
)) as f:
    ProxyFactory = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/gnosis-safe",
    "GnosisSafe.json",
)) as f:
    GnosisSafe = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/open-zeppelin",
    "TokenTimelock.json",
)) as f:
    TokenTimelock = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/open-zeppelin-upgrades",
    "ProxyAdmin.json",
)) as f:
    ProxyAdmin = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/multicall",
    "Multicall.json",
)) as f:
    Multicall = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/digg/Mock.json",
)) as f:
    Mock = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/open-zeppelin-upgrades",
    "AdminUpgradeabilityProxy.json",
)) as f:
    AdminUpgradeabilityProxy = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/uniswap",
    "UniswapV2Pair.json",
)) as f:
    UniswapV2Pair = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/uniswap",
    "UniswapV2Factory.json",
)) as f:
    UniswapV2Factory = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/uniswap",
    "UniswapV2Router02.json",
)) as f:
    UniswapV2Router = json.load(f)

with open(os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "dependency-artifacts/wbtc",
    "wbtc.json",
)) as f:
    wbtc = json.load(f)

artifacts = DotMap(
    aragon=DotMap(
        Agent=Agent,
        CompanyTemplate=CompanyTemplate,
        Vault=Vault,
        Voting=Voting,
        Finance=Finance,
        TokenManager=TokenManager,
        MiniMeToken=MiniMeToken,
    ),
    gnosis_safe=DotMap(
        MasterCopy=MasterCopy, ProxyFactory=ProxyFactory, GnosisSafe=GnosisSafe
    ),
    digg=DotMap(
        Mock=Mock,
    ),
    open_zeppelin=DotMap(
        ProxyAdmin=ProxyAdmin,
        AdminUpgradeabilityProxy=AdminUpgradeabilityProxy,
        TokenTimelock=TokenTimelock,
    ),
    uniswap=DotMap(
        UniswapV2Factory=UniswapV2Factory,
        UniswapV2Router=UniswapV2Router,
        UniswapV2Pair=UniswapV2Pair,
    ),
    multicall=DotMap(multicall={"abi": Multicall}),
    wbtc=DotMap(wbtc={"abi": wbtc}),
)



from badger_utils.registry.artifacts import artifacts
from badger_utils.registry.chain_registry import ChainRegistry
from badger_utils.registry.on_chain_registries import chain_registries
from badger_utils.registry.registries import registries
from badger_utils.registry.registries import registry


__all__ = [
    "registries",
    "registry",
    "artifacts",
    "chain_registries",
]



from brownie import Wei
from brownie.network import gas_price
from brownie.network.gas.strategies import ExponentialScalingStrategy
from brownie.network.gas.strategies import SimpleGasStrategy
from web3 import Web3

from badger_utils.constants import BSC_STATIC_PRICE
from badger_utils.constants import EXPONENTIAL_SCALING_CONFIG
from badger_utils.gas_utils.analyze_gas import analyze_gas
from badger_utils.gas_utils.gas_now_compatible_strategy import GasNowCompatibleStrategy
from badger_utils.network_manager import network_manager


class StaticGasStrategy(SimpleGasStrategy):
    def __init__(self, price) -> None:
        self.price = price
        super().__init__()

    def get_gas_price(self) -> int:
        return self.price


class GasStrategies:
    def __init__(self):
        self.standard = GasNowCompatibleStrategy("standard")
        self.fast = GasNowCompatibleStrategy("fast")
        self.rapid = GasNowCompatibleStrategy("rapid")
        self.bsc_static = StaticGasStrategy(BSC_STATIC_PRICE)
        self.analyzed = analyze_gas({"timeframe": "minutes", "periods": 15})

        self.exponential_scaling = ExponentialScalingStrategy(
            initial_gas_price=self.standard.get_gas_price(),
            max_gas_price=Wei(EXPONENTIAL_SCALING_CONFIG["max_gas_price"]),
            time_duration=120,
        )

        self.exponential_scaling_fast = ExponentialScalingStrategy(
            initial_gas_price=self.fast.get_gas_price(),
            max_gas_price=Wei(EXPONENTIAL_SCALING_CONFIG["max_gas_price"]),
            time_duration=60,
        )

    def set_default(self, strategy) -> None:
        gas_price(strategy)

    def gas_cost(self, gas_estimate: int) -> int:
        """
        Total gas cost of estimate in wei
        """
        return Web3.toWei(
            Web3.fromWei(self.fast.get_gas_price(), "gwei") * gas_estimate, "gwei"
        )

    def set_default_for_active_chain(self) -> None:
        chain = network_manager.get_active_network()
        if chain == "eth":
            self.set_default(self.exponential_scaling)
        elif chain == "bsc":
            self.set_default(self.bsc_static)

    def optimal_price(self) -> int:
        return min(self.fast.get_gas_price(), self.analyzed.mode)



from brownie.network.contract import Contract
from brownie.network.account import Account


class Balances:
    def __init__(self):
        self.balances = {}

    def set(self, token: Contract, account: Account, value: int):
        if token.address not in self.balances:
            self.balances[token.address] = {}
        self.balances[token.address][account.address] = value

    def get(self, token, account):
        return self.balances[token][account]



from badger_utils.systems.aave_system import AaveSystem
from badger_utils.systems.compound_system import CompoundSystem
from badger_utils.systems.sushiswap_system import SushiswapSystem
from badger_utils.systems.token_system import TokenSystem
from badger_utils.systems.yearn_system import YearnSystem

__all__ = [
    "AaveSystem",
    "CompoundSystem",
    "SushiswapSystem",
    "TokenSystem",
    "YearnSystem",
]



import os
from time import time
from typing import Dict
from typing import List
from typing import Optional

import numpy as np
from dotmap import DotMap
from elasticsearch import Elasticsearch

from badger_utils.constants import HISTORICAL_MAINNET_ELASTIC_URL
from badger_utils.constants import NUMBER_OF_BINS


def is_outlier(points: np.array, thresh=3.5) -> bool:
    """
    Returns a boolean array with True if points are outliers and False
    otherwise.

    Parameters:
    -----------
        points : A numobservations by numdimensions array of observations
        thresh : The modified z-score to use as a threshold. Observations with
            a modified z-score (based on the median absolute deviation) greater
            than this value will be classified as outliers.

    Returns:
    --------
        mask : A numobservations-length boolean array.

    References:
    ----------
        Boris Iglewicz and David Hoaglin (1993), "Volume 16: How to Detect and
        Handle Outliers", The ASQC Basic References in Quality Control:
        Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.
    """
    if len(points.shape) == 1:
        points = points[:, None]
    median = np.median(points, axis=0)
    diff = np.sum((points - median) ** 2, axis=-1)
    diff = np.sqrt(diff)
    med_abs_deviation = np.median(diff)

    modified_z_score = 0.6745 * diff / med_abs_deviation

    return modified_z_score > thresh


def fetch_gas_hour(network: str, hours=24) -> List[float]:
    """
    Fetch average hourly gas prices over the last specified hours
    """
    es = Elasticsearch(
        hosts=[network],
        http_auth=(os.environ.get("ANYBLOCK_EMAIL"), os.environ.get("ANYBLOCK_KEY")), timeout=180
    )
    now = int(time())
    seconds = hours * 3600
    data = es.search(
        index="tx",
        doc_type="tx",
        body={
            "_source": ["timestamp", "gasPrice.num"],
            "query": {
                "bool": {
                    "must": [
                        {"range": {"timestamp": {"gte": now - seconds, "lte": now}}}
                    ]
                }
            },
            "aggs": {
                "hour_bucket": {
                    "date_histogram": {
                        "field": "timestamp",
                        "interval": "1H",
                        "format": "yyyy-MM-dd hh:mm:ss",
                    },
                    "aggs": {"avgGasHour": {"avg": {"field": "gasPrice.num"}}},
                }
            },
        },
    )
    return [
        x["avgGasHour"]["value"]
        for x in data["aggregations"]["hour_bucket"]["buckets"]
        if x["avgGasHour"]["value"]
    ]


# fetch average gas prices per minute over the last specified minutes
def fetch_gas_min(network: str, minutes=60) -> List[float]:
    es = Elasticsearch(
        hosts=[network],
        http_auth=(os.environ.get("ANYBLOCK_EMAIL"), os.environ.get("ANYBLOCK_KEY")), timeout=180
    )
    now = int(time())
    seconds = minutes * 60
    data = es.search(
        index="tx",
        doc_type="tx",
        body={
            "_source": ["timestamp", "gasPrice.num"],
            "query": {
                "bool": {
                    "must": [
                        {"range": {"timestamp": {"gte": now - seconds, "lte": now}}}
                    ]
                }
            },
            "aggs": {
                "minute_bucket": {
                    "date_histogram": {
                        "field": "timestamp",
                        "interval": "1m",
                        "format": "yyyy-MM-dd hh:mm",
                    },
                    "aggs": {"avgGasMin": {"avg": {"field": "gasPrice.num"}}},
                }
            },
        },
    )
    return [
        x["avgGasMin"]["value"]
        for x in data["aggregations"]["minute_bucket"]["buckets"]
        if x["avgGasMin"]["value"]
    ]


def analyze_gas(options: Optional[Dict] = None) -> DotMap:
    if not options:
        options = {
            "timeframe": "minutes", "periods": 60
        }
    if not os.environ.get("ANYBLOCK_EMAIL") or not os.environ.get("ANYBLOCK_KEY"):
        # Could not fetch historical gas data
        return DotMap(
            mode=999999999999999999, median=999999999999999999, std=999999999999999999
        )

    # fetch data
    if options["timeframe"] == "minutes":
        gas_data = fetch_gas_min(HISTORICAL_MAINNET_ELASTIC_URL, options["periods"])
    else:
        gas_data = fetch_gas_hour(HISTORICAL_MAINNET_ELASTIC_URL, options["periods"])
    gas_data = np.array(gas_data)

    # remove outliers
    filtered_gas_data = gas_data[~is_outlier(gas_data)]

    # Create histogram
    counts, bins = np.histogram(filtered_gas_data, bins=NUMBER_OF_BINS)

    # Find most common gas price
    biggest_bin = 0
    biggest_index = 0
    for i, x in enumerate(counts):
        if x > biggest_bin:
            biggest_bin = x
            biggest_index = i

    midpoint = (bins[biggest_index] + bins[biggest_index + 1]) / 2

    if int(midpoint) == 0:
        # Could not fetch historical gas data
        return DotMap(
            mode=999999999999999999, median=999999999999999999, std=999999999999999999
        )

    # standard deviation
    standard_dev = np.std(filtered_gas_data, dtype=np.float64)

    # average
    median = np.median(filtered_gas_data, axis=0)

    return DotMap(mode=int(midpoint), median=int(median), std=int(standard_dev))



from brownie import interface


class CompoundSystem:
    def __init__(self, registry):
        self.registry = registry
        self.comptroller = interface.IComptroller(registry.compound.comptroller)

    def ctoken(self, name: str):
        if name not in self.registry.compound.cTokens:
            raise Exception(f"No cToken found for key {name}")
        return interface.ICToken(self.registry.compound.cTokens[name])



from dotmap import DotMap

from badger_utils.constants import WhaleRegistryAction
from badger_utils.registry.chain_registry import ChainRegistry

gnosis_safe_registry = DotMap(
    addresses=DotMap(
        proxyFactory="0x76E2cFc1F5Fa8F6a5b3fC4c8F4788F0116861F9B",
        masterCopy="0x34CfAC646f301356fAa8B21e94227e3583Fe3F5F",
    ),
)
pancake_registry = DotMap(
    cake="0x0e09fabb73bd3ade0a17ecc321fd13a19e81ce82",
    symbol="Cake",
    syrup="0x009cF7bC57584b7998236eff51b98A168DceA9B0",
    masterChef="0x73feaa1eE314F8c655E354234017bE2193C9E24E",
    factoryV2="0xBCfCcbde45cE874adCB698cC183deBcF17952812",
    routerV2="0x05fF2B0DB69458A0750badebc4f9e13aDd608C7F",
    smartChefs="0xe4dD0C50fb314A8B2e84D211546F5B57eDd7c2b9",
    chefPairs=DotMap(
        bnbBtcb="0x7561EEe90e24F3b348E1087A005F78B4c8453524",
        bBadgerBtcb="0x10f461ceac7a17f59e249954db0784d42eff5db5",
        bDiggBtcb="0xE1E33459505bB3763843a426F7Fd9933418184ae",
    ),
    chefPids=DotMap(
        bnbBtcb=15,
        bBadgerBtcb=0,
        bDiggBtcb=104,
    ),
)
multicall_registry = DotMap(
    multicall="0xE1dDc30f691CA671518090931e3bFC1184BFa4Aa",
)
sushi_registry = DotMap(
    sushiToken="",
    xsushiToken="",
    sushiChef="",
    router="",
    factory="",
    lpTokens=DotMap(
        sushiBadgerWBtc="",
        sushiWbtcWeth="",
    ),
    pids=DotMap(sushiBadgerWBtc=0, sushiEthWBtc=0),
)
token_registry = DotMap(
    bnb="0xbb4CdB9CBd36B01bD1cBaEBF2De08d9173bc095c",
    btcb="0x7130d2A12B9BCbFAe4f2634d864A1Ee1Ce3Ead9c",
    usdc="0x8ac76a51cc950d9822d68b83fe1ad97b32cd580d",
    bBadger="0x1F7216fdB338247512Ec99715587bb97BBf96eae",
    bDigg="0x5986D5c77c65e5801a5cAa4fAE80089f870A71dA",
)

bsc_registry = ChainRegistry(
    sushiswap=sushi_registry,
    sushi=sushi_registry,
    multicall=multicall_registry,
    pancake=pancake_registry,
    tokens=token_registry,
)

bsc_registry.whales = DotMap(
    bnbBtcb=DotMap(
        whale=bsc_registry.pancake.masterChef,
        token=bsc_registry.pancake.chefPairs.bnbBtcb,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bDiggBtcb=DotMap(
        whale=bsc_registry.pancake.masterChef,
        token=bsc_registry.pancake.chefPairs.bDiggBtcb,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bBadgerBtcb=DotMap(
        whale=bsc_registry.pancake.masterChef,
        token=bsc_registry.pancake.chefPairs.bBadgerBtcb,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    # Need these for `test_simulation.py` on bsc
    bnb=DotMap(
        whale="0xd7D069493685A581d27824Fc46EdA46B7EfC0063",
        token=bsc_registry.tokens.bnb,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    btcb=DotMap(
        whale="0x631Fc1EA2270e98fbD9D92658eCe0F5a269Aa161",
        token=bsc_registry.tokens.btcb,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bBadger=DotMap(
        whale="0x5a58609da96469e9def3fe344bc39b00d18eb9a5",
        token=bsc_registry.tokens.bBadger,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    bDigg=DotMap(
        whale="0x81d776c90c89b8d51e9497d58338933127e2fa80",
        token=bsc_registry.tokens.bDigg,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
    pancakebBadgerbtcb=DotMap(
        whale="0x2A842e01724F10d093aE8a46A01e66DbCf3C7373",
        token=bsc_registry.tokens.bDigg,
        action=WhaleRegistryAction.DISTRIBUTE_FROM_CONTRACT,
    ),
)



import os
from threading import Thread
from time import sleep
from time import time
from typing import Optional

import requests
from brownie.network.account import Account


class TxTimer:
    """
    Starts a timer in a new thread that will send a message to the specified webhook
    if the timer is not turned off before `time_threshold`.

    Usage:
    - call the start_timer method before sending a transaction
    - call the end_timer method on the next line after sending the transaction

    Example:
    tx_timer.start_timer(overrides['from'], 'Harvest')
    tx_timer.end_timer()
    """

    def __init__(self, time_threshold=1200, timer_tick=1) -> None:
        """
        :param time_threshold: time in seconds to wait for sending alert
        :param timer_tick: how long to sleep thread
            before resuming loop to see if alert should be sent

        Attributes:
        waiting: whether or not there is a tx currently waiting
        sender: account that tx was sent from, used in alert message
        webhook: webhook url from .env
        tx_type: Harvest, Tend, etc. included in alert message
        """

        self.time_threshold = time_threshold
        self.timer_tick = timer_tick
        self.waiting = False
        self._sender = None
        self.webhook = os.environ.get("TX_TIMER_WEBHOOK")
        self.tx_type = ""
        self.thread = None

    @property
    def sender(self) -> Optional[Account]:
        return self._sender

    @sender.setter
    def sender(self, account: Account) -> None:
        self._sender = account

    def alert(self, msg: str) -> None:
        if self.webhook:
            requests.post(self.webhook, {"content": msg})

    def track_tx(self) -> None:
        """
        Run a loop until the timer is ended or `time_threshold` is hit. If the threshold is hit
        then send an alert to the webhook
        """

        start = time()
        self.waiting = True
        while self.waiting:
            sleep(self.timer_tick)
            if time() - start >= self.time_threshold:
                if self.tx_type:
                    msg = (
                        f"🕔 * {self.tx_type} * - tx sent from {str(self.sender.address)}"
                        f" has exceeded threshold of {str(self.time_threshold)} seconds"
                    )
                else:
                    msg = (
                        f"🕔 tx sent from {str(self.sender.address)}  has exceeded threshold of "
                        f"{str(self.time_threshold)} seconds"
                    )
                self.alert(msg)
                self.waiting = False
                self.sender = None
                self.tx_type = ""

    def prepare_timer(self, sender: Account, tx_type: str) -> None:
        self.sender = sender
        self.tx_type = tx_type

    def start_timer(self, sender: Account, tx_type: str) -> None:
        self.prepare_timer(sender, tx_type)
        self.thread = Thread(target=self.track_tx)
        self.thread.start()

    def end_timer(self) -> None:
        self.waiting = False
        self.sender = None
        self.tx_type = ""


tx_timer = TxTimer()



from brownie import interface


class YearnSystem:
    def __init__(self, chain_registry):
        self.chain_registry = chain_registry

    def registry(self):
        return interface.RegistryAPI(self.chain_registry.yearn.registry)

    def experimental_vault_by_key(self, key):
        if key not in self.chain_registry.yearn.experimental_vaults:
            raise Exception("Token with key {} not found in registry".format(key))

        address = self.chain_registry.yearn.experimental_vaults[key]
        return self.vault_by_address(address)

    def vault_by_address(self, address):
        return interface.VaultAPI(address)



from typing import Optional

from brownie import interface
from brownie import web3


def approx(actual: int, expected: int, percentage_threshold: int) -> bool:
    diff = int(abs(actual - expected))
    if diff == 0:
        return True
    return diff < (actual * percentage_threshold // 100)


def val(amount=0, decimals=18, token: Optional[str] = None):
    # If no token specified, use decimals
    if token:
        decimals = interface.IERC20(token).decimals()

    return "{:,.18f}".format(amount / 10 ** decimals)


def is_address_eoa(address: str) -> bool:
    # If address is a contract, `get_code` will return hashed contract bytecode,
    # otherwise - empty bytestring
    return web3.eth.get_code(address) == b''



import pathlib

from setuptools import setup
from setuptools import find_packages


with open("requirements.txt", "r") as f:
    requirements = list(map(str.strip, f.read().split("\n")))[:-1]

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()


setup(
    name="badger-utils",
    install_requires=requirements,
    author="Andrii Kulikov",
    author_email="blaynemono@gmail.com",
    description="Badger utils and shared code",
    long_description=README,
    keywords=["badger-utils"],
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    version="0.2.2",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
    ],
    license="MIT",
    url="https://github.com/SHAKOTN/badger-utils",
    python_requires=">=3.7,<4",
)



from enum import Enum

from brownie import Wei
from brownie.network import web3

AddressZero = "0x0000000000000000000000000000000000000000"
MaxUint256 = str(int(2 ** 256 - 1))
EmptyBytes32 = "0x0000000000000000000000000000000000000000000000000000000000000000"

DEFAULT_ADMIN_ROLE = (
    "0x0000000000000000000000000000000000000000000000000000000000000000"
)

TOKEN_LOCKER_ROLE = web3.keccak(text="TOKEN_LOCKER_ROLE").hex()
ROOT_UPDATER_ROLE = web3.keccak(text="ROOT_UPDATER_ROLE").hex()
GUARDIAN_ROLE = web3.keccak(text="GUARDIAN_ROLE").hex()
APPROVED_STAKER_ROLE = web3.keccak(text="APPROVED_STAKER_ROLE").hex()
PAUSER_ROLE = web3.keccak(text="PAUSER_ROLE").hex()
UNPAUSER_ROLE = web3.keccak(text="UNPAUSER_ROLE").hex()
DISTRIBUTOR_ROLE = web3.keccak(text="DISTRIBUTOR_ROLE").hex()
ROOT_PROPOSER_ROLE = web3.keccak(text="ROOT_PROPOSER_ROLE").hex()
ROOT_VALIDATOR_ROLE = web3.keccak(text="ROOT_VALIDATOR_ROLE").hex()
APPROVED_ACCOUNT_ROLE = web3.keccak(text="APPROVED_ACCOUNT_ROLE").hex()

DIGG = "0x798D1bE841a82a273720CE31c822C61a67a601C3"
BADGER = "0x3472A5A71965499acd81997a54BBA8D852C6E53d"
BBADGER = "0x19d97d8fa813ee2f51ad4b4e04ea08baf4dffc28"
SPELL = "0x090185f2135308bad17527004364ebcc2d37e5f6"
FARM = "0xa0246c9032bC3A600820415aE600c6388619A14D"
XSUSHI = "0x8798249c2E607446EfB7Ad49eC89dD1865Ff4272"
DFD = "0x20c36f062a31865bED8a5B1e512D9a1A20AA333A"
BCVXCRV = "0x2B5455aac8d64C14786c3a29858E43b5945819C0"
BCVX = "0x53C8E199eb2Cb7c01543C137078a038937a68E40"
PNT = "0x89Ab32156e46F46D02ade3FEcbe5Fc4243B9AAeD"
BOR = "0x3c9d6c1C73b31c837832c72E04D3152f051fc1A9"

TOKENS_TO_CHECK = {
    "Badger": BADGER,
    "Digg": DIGG,
    "Farm": FARM,
    "xSushi": XSUSHI,
    "Dfd": DFD,
    "bCvxCrv": BCVXCRV,
    "bCvx": BCVX,
    "Pnt": PNT,
    "Bor": BOR,
}

BADGER_TREE = "0x660802Fc641b154aBA66a62137e71f331B6d787A"

PEAK_ADDRESSES = [
    "0x825218beD8BE0B30be39475755AceE0250C50627",
    "0x41671BA1abcbA387b9b2B752c205e22e916BE6e3",
]

MAX_BOOST = 3
DIGG_SETTS = ["native.uniDiggWbtc", "native.sushiDiggWbtc", "native.digg"]
BADGER_SETTS = ["native.badger", "native.uniBadgerWbtc", "native.sushiBadgerWbtc"]
NATIVE_DIGG_SETTS = ["native.uniDiggWbtc", "native.sushiDiggWbtc"]

NON_NATIVE_SETTS = [
    "native.renCrv",
    "native.sbtcCrv",
    "native.tbtcCrv",
    "native.sushiWbtcEth",
    "harvest.renCrv",
    "yearn.wbtc",
    "experimental.sushiIBbtcWbtc",
    "native.hbtcCrv",
    "native.pbtcCrv",
    "native.obtcCrv",
    "native.bbtcCrv",
    "native.tricrypto",
    "native.cvxCrv",
    "native.cvx",
]

NO_GEYSERS = [
    "native.digg",
    "experimental.sushiIBbtcWbtc",
    "experimental.digg",
    "native.hbtcCrv",
    "native.pbtcCrv",
    "native.obtcCrv",
    "native.bbtcCrv",
    "native.tricrypto",
    "native.cvxCrv",
    "native.cvx",
]

SETT_BOOST_RATIOS = {
    "native.uniDiggWbtc": 0.5,
    "native.sushiDiggWbtc": 0.5,
    "native.uniBadgerWbtc": 0.5,
    "native.sushiBadgerWbtc": 0.5,
    "native.badger": 1,
    "native.digg": 1,
    "native.renCrv": 1,
    "native.sbtcCrv": 1,
    "native.tbtcCrv": 1,
    "harvest.renCrv": 1,
    "native.sushiWbtcEth": 1,
    "yearn.wbtc": 1,
    "experimental.sushiIBbtcWbtc": 1,
    "native.hbtcCrv": 1,
    "native.pbtcCrv": 1,
    "native.obtcCrv": 1,
    "native.bbtcCrv": 1,
    "native.tricrypto": 1,
    "native.cvxCrv": 0.1,
    "native.cvx": 0.1,
}

CONVEX_SETTS = ["native.hbtcCrv", "native.pbtcCrv", "native.obtcCrv", "native.bbtcCrv"]


class WhaleRegistryAction(Enum):
    DISTRIBUTE_FROM_EOA = (0,)
    DISTRIBUTE_FROM_CONTRACT = (1,)
    POPULATE_NEW_SUSHI_LP = 2


class ContractSystems(Enum):
    ARAGON = ("aragon",)
    GNOSIS_SAFE = ("gnosis-safe",)
    OPEN_ZEPPELIN = ("open-zeppelin",)
    UNISWAP = ("uniswap",)
    SUSHISWAP = ("sushiswap",)
    MULTICALL = ("multicall",)
    PICKLE = ("pickle",)
    HARVEST = ("harvest",)
    CURVE = ("curve",)
    CHAINLINK = ("chainlink",)
    TOKENS = "tokens"


BADGER_REGISTRY_ADDRESS = "0xFda7eB6f8b7a9e9fCFd348042ae675d1d652454f"  # Multichain BadgerRegistry

ETHEREUM_NETWORK = "eth"
BINANCE_NETWORK = "bsc"
POLYGON_NETWORK = "polygon"
ARBITRUM_NETWORK = "arbitrum"
DEVELOPMENT_NETWORK = "development"

SUPPORTED_NETWORKS = [ETHEREUM_NETWORK, BINANCE_NETWORK, POLYGON_NETWORK, ARBITRUM_NETWORK]


EXPONENTIAL_SCALING_CONFIG = {
    "initial_gas_price": "100 gwei",
    "max_gas_price": "1000 gwei",
}

BSC_STATIC_PRICE = Wei("10 gwei")
HISTORICAL_MAINNET_ELASTIC_URL = "https://api.anyblock.tools/ethereum/ethereum/mainnet/es/"

NUMBER_OF_BINS = 60  # number of bins for histogram

TARGET_TOKENS = [
    "0x3472A5A71965499acd81997a54BBA8D852C6E53d",
    "0x19D97D8fA813EE2f51aD4B4e04EA08bAf4DFfC28",
    "0xAf5A1DECfa95BAF63E0084a35c62592B774A2A87",
    "0xcD7989894bc033581532D2cd88Da5db0A4b12859",
    "0xe86204c4eddd2f70ee00ead6805f917671f56c52",
    "0x075b1bb99792c9E1041bA13afEf80C91a1e70fB3",
    "0xd04c48A53c111300aD41190D63681ed3dAd998eC",
    "0x49849C98ae39Fff122806C06791Fa73784FB3675",
    "0x6dEf55d2e18486B9dDfaA075bc4e4EE0B28c1545",
    "0x64eda51d3Ad40D56b9dFc5554E06F94e1Dd786Fd",
    "0xb9D076fDe463dbc9f915E5392F807315Bf940334",
    "0xb19059ebb43466C323583928285a49f558E572Fd",
    "0xDE5331AC4B3630f94853Ff322B66407e0D6331E8",
    "0x2fE94ea3d5d4a175184081439753DE15AeF9d614",
    "0x410e3E86ef427e30B9235497143881f717d93c2A",
    "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599",
    "0x110492b31c59716AC47337E616804E3E3AdC0b4a",
    "0xCEfF51756c56CeFFCA006cD410B03FFC46dd3a58",
    "0x758A43EE2BFf8230eeb784879CdcFF4828F2544D",
    "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48",
    "0x798D1bE841a82a273720CE31c822C61a67a601C3",
    "0xeb4c2781e4eba804ce9a9803c67d0893436bb27d",
    "0xcA3d75aC011BF5aD07a98d02f18225F9bD9A6BDF",
    "0x4e3fbd56cd56c3e72c1403e103b45db9da5b9d2b",
    "0x62b9c7356a2dc64a1969e19c23e4f579f9810aa7",
    "0xFbdCA68601f835b27790D98bbb8eC7f05FDEaA9B",
]



from badger_utils.token_utils.balances import Balances
from badger_utils.token_utils.utils import distribute_from_whales
from badger_utils.token_utils.utils import get_token_balances


__all__ = [
    "Balances",
    "distribute_from_whales",
    "get_token_balances",
]



import re
import sys
from typing import Optional

from brownie import network

from badger_utils.constants import ARBITRUM_NETWORK
from badger_utils.constants import DEVELOPMENT_NETWORK
from badger_utils.constants import BINANCE_NETWORK
from badger_utils.constants import ETHEREUM_NETWORK
from badger_utils.constants import POLYGON_NETWORK


class NetworkManager:
    @staticmethod
    def network_name(raw_network_name: str) -> Optional[str]:
        if re.match(r"^mainnet", raw_network_name):
            return ETHEREUM_NETWORK
        if re.match(r"(?:bsc|binance)", raw_network_name):
            return BINANCE_NETWORK
        if re.match(r"^polygon", raw_network_name):
            return POLYGON_NETWORK
        if re.match(r"^arbitrum", raw_network_name):
            return ARBITRUM_NETWORK
        if re.match(r"^development", raw_network_name):
            return DEVELOPMENT_NETWORK
        return None

    @staticmethod
    def get_active_network() -> str:
        active_network = network.show_active()

        if active_network is None:
            if "--network" not in sys.argv:
                name = "eth"
            else:
                network_idx = sys.argv.index("--network")
                name = NetworkManager.network_name(sys.argv[network_idx + 1])
        else:
            name = NetworkManager.network_name(active_network)

        if not name:
            raise Exception(f"Chain ID {active_network} not recognized")

        return name

    @staticmethod
    def get_active_network_badger_deploy() -> str:
        active_network = NetworkManager.get_active_network()
        if active_network == ETHEREUM_NETWORK:
            return "deploy-final.json"
        elif active_network == BINANCE_NETWORK:
            return "badger-deploy-bsc.json"
        else:
            raise Exception(f"No badger deploy file registered for network {active_network}")


network_manager = NetworkManager()



from badger_utils.network_manager import network_manager
from badger_utils.registry.chain_registry import ChainRegistry


class BaseRegistry:
    def __init__(self):
        self.registries = {}

    def get_registry(self, chain: str):
        return self.registries[chain]

    def has_registry(self, chain: str) -> bool:
        return chain in self.registries.keys()

    def get_active_chain_registry(self) -> ChainRegistry:
        network_id = network_manager.get_active_network()
        return self.get_registry(network_id)
